/* Process model C form file: UniApp.pr.c */
/* Portions of this file Copyright (C) MIL 3, Inc. 1992 */



/* OPNET system definitions */
#include <opnet.h>
#include "UniApp.pr.h"
FSM_EXT_DECS



/* Header block */
/*
** This are the conditions for transitions to got to different states from the idle block
*/
#define	PK_ARR		op_intrpt_type () == OPC_INTRPT_STRM
#define	SELF_INT1	op_intrpt_type () == OPC_INTRPT_SELF  && op_intrpt_code () == 0
#define END_SIM		op_intrpt_type () ==  OPC_INTRPT_ENDSIM

/*
** Different probes that are used to meausure the number of data packets sent, no. of data
** packets received by this module (or process)
*/
#define CUMM_PKT_SENT_STAT  0
#define CUMM_PKT_RECV_STAT  1
#define AVG_PKT_RECV_STAT   2

#define send_data_intr      0
#define IP_STREAM           0

#define Meas_Intvl          1

#define UDP_PROTOCOL_ID     20

#define ON                  1
#define OFF                 0

#define VL                  if(OFF)
#define UADL                if(OFF)

#define PROBE_ON            if(ON)
#define PROBE_OFF           if(OFF)

typedef struct 
{
	int NetNum;
	int NodeNum;
}IPAddress;

typedef struct 
{
	char Message[500];
}App_Data;



/* State variable definitions */
typedef struct
	{
	FSM_SYS_STATE
	Objid                   		sv_my_objid;
	Objid                   		sv_ip_objid;
	double                  		sv_data_iat;
	double                  		sv_print_time;
	double                  		sv_NoOfPktsRecv;
	double                  		sv_NoOfPktsSent;
	int                     		sv_MAX_NO_NETS;
	int                     		sv_MAX_NO_NODES;
	int                     		sv_DataSendInitTime;
	Distribution*           		sv_net_dist;
	Distribution*           		sv_node_dist;
	IPAddress               		sv_LocalAddress;
	} UniApp_state;

#define pr_state_ptr            		((UniApp_state*) SimI_Mod_State_Ptr)
#define my_objid                		pr_state_ptr->sv_my_objid
#define ip_objid                		pr_state_ptr->sv_ip_objid
#define data_iat                		pr_state_ptr->sv_data_iat
#define print_time              		pr_state_ptr->sv_print_time
#define NoOfPktsRecv            		pr_state_ptr->sv_NoOfPktsRecv
#define NoOfPktsSent            		pr_state_ptr->sv_NoOfPktsSent
#define MAX_NO_NETS             		pr_state_ptr->sv_MAX_NO_NETS
#define MAX_NO_NODES            		pr_state_ptr->sv_MAX_NO_NODES
#define DataSendInitTime        		pr_state_ptr->sv_DataSendInitTime
#define net_dist                		pr_state_ptr->sv_net_dist
#define node_dist               		pr_state_ptr->sv_node_dist
#define LocalAddress            		pr_state_ptr->sv_LocalAddress




/* Process model interrupt handling procedure */



void
UniApp ()
	{
	double thru, thru1, tmpDataRec, tmpDataSend, tmp_sim_time, thru2, thru3;
	Packet*  pkptr;
	int in_strm,da1,da2,da3;
	FILE*  fil;
	


	FSM_ENTER (UniApp)

	FSM_BLOCK_SWITCH
		{
		/*---------------------------------------------------------*/
		/** state (init) enter executives **/
		FSM_STATE_ENTER_FORCED (0, state0_enter_exec, "init")
			{
			/*******************************************************************************************
			** Process Model : Uniapp                                                                 **
			** State         : init                                                                   **
			**                                                                                        **
			** This block intializes the variables that will be used in this process model, loads the **
			** distributions that are used to select the destination address to which the data packet **
			** is sent. It also  gets the object id of the IP module and  the IP address of this host **
			** from that IP module. This also schedules an interrupt for sending data after some init **
			** time. This  init time is given  so that this unicast application doesn't send any data **
			** before all the routing tables are initialized.                                         **
			**                                                                                        **
			*******************************************************************************************/
			
			/*
			** Get the present module's  object ID and load the attributes from  the configuaration file
			*/
			my_objid = op_id_self();
			op_ima_obj_attr_get( my_objid, "MaxNets", &MAX_NO_NETS );
			op_ima_obj_attr_get( my_objid, "MaxNodes", &MAX_NO_NODES );
			op_ima_obj_attr_get( my_objid, "InitTime", &DataSendInitTime );
			op_ima_obj_attr_get( my_objid,"DATA_IAT", &data_iat);
			
			/* 
			** Get the IP module's Object Id  and use this object id to  get the IP address  of the host
			*/
			ip_objid = op_topo_assoc (my_objid, OPC_TOPO_ASSOC_OUT, OPC_OBJMTYPE_MODULE, 0);
			op_ima_obj_attr_get( ip_objid, "net_num0", &LocalAddress.NetNum );
			op_ima_obj_attr_get( ip_objid, "node_num0", &LocalAddress.NodeNum );
			
			/*
			** Load the distribution pointers for selecting the destination address for the data packet 
			** that is being sent from this state
			*/
			net_dist = op_dist_load("uniform_int", 1, MAX_NO_NETS );
			node_dist = op_dist_load("uniform_int", 1, MAX_NO_NODES );
			
			/*
			** Schedule the interrupt to send the data packets 
			*/
			op_intrpt_schedule_self( (op_sim_time() + data_iat + DataSendInitTime), send_data_intr );
			}


		/** state (init) exit executives **/
		FSM_STATE_EXIT_FORCED (0, state0_exit_exec, "init")
			{
			}


		/** state (init) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (idle) enter executives **/
		FSM_STATE_ENTER_UNFORCED (1, state1_enter_exec, "idle")
			{
			}


		/** blocking after enter executives of unforced state. **/
		FSM_EXIT (3,UniApp)


		/** state (idle) exit executives **/
		FSM_STATE_EXIT_UNFORCED (1, state1_exit_exec, "idle")
			{
			}


		/** state (idle) transition processing **/
		FSM_INIT_COND (PK_ARR)
		FSM_TEST_COND (SELF_INT1)
		FSM_TEST_LOGIC ("idle")

		FSM_TRANSIT_SWITCH
			{
			FSM_CASE_TRANSIT (0, 2, state2_enter_exec, ;)
			FSM_CASE_TRANSIT (1, 3, state3_enter_exec, ;)
			}
		/*---------------------------------------------------------*/



		/** state (pk_arr) enter executives **/
		FSM_STATE_ENTER_FORCED (2, state2_enter_exec, "pk_arr")
			{
			
			/*****************************************************************************************************
			** Process Model : Uniapp                                                                           **
			** State         : pk_arr                                                                           **
			**                                                                                                  **
			** This is a port for all the traffic(packets  with protocol ID as 20) that was destinated for this **
			** host. The state upon receiving the packet updates the counter which keeps track of the number of **
			** packets received  by this host. It writes this statistics into a probe file which can be plotted **
			** once the simulation is completed. It destroys the packets that are received.                     **
			**                                                                                                  **
			*****************************************************************************************************/
			
			pkptr  = op_pk_get (op_intrpt_strm ());
			
			/*
			** Probe to measure the number of data packets received by this UDP port
			*/
			PROBE_ON
			{
				NoOfPktsRecv++;
				op_stat_local_write( CUMM_PKT_RECV_STAT, NoOfPktsRecv );
			}
			PROBE_OFF
			{
				tmpDataRec++;
				if(op_sim_time() > tmp_sim_time + Meas_Intvl )
				{
					thru1 = tmpDataRec/Meas_Intvl;
					op_stat_local_write( AVG_PKT_RECV_STAT, thru1);
					tmpDataRec = 0;
					tmp_sim_time = tmp_sim_time + Meas_Intvl;
				}
			}
			op_pk_destroy (pkptr);
			}


		/** state (pk_arr) exit executives **/
		FSM_STATE_EXIT_FORCED (2, state2_exit_exec, "pk_arr")
			{
			}


		/** state (pk_arr) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (data) enter executives **/
		FSM_STATE_ENTER_FORCED (3, state3_enter_exec, "data")
			{
			/*******************************************************************************************
			** Process Model : Uniapp                                                                 **
			** State         : data                                                                   **
			**                                                                                        **
			** This block selects a destination address randomly(from the uniform distributions) from **
			** the networks that are present into the model. It packs the data packet encapsulates it **
			** into the IP datagram and sends the packet to the IP layer. The message is of 500 bytes **
			** length. This block also keeps track of the no. of packets sent by this host and writes **
			** this  statistic into a probe file that can  plotted through the analysis tool. It also **
			** schedules an interrput for sending the next data packet.                               **
			**                                                                                        **
			*******************************************************************************************/
			
			App_Data DataSent;
			Packet* ip_pkptr;
			int dest_net, dest_node, data_len;
			
			/*
			** Get the destination address from the uniform distributions
			*/
			dest_net = op_dist_outcome( net_dist );
			dest_node = op_dist_outcome( node_dist );
			
			if( dest_net == 5 )
				dest_net++;
			sprintf( DataSent.Message, "This was data sending to the host %d.%d by %d.%d \n", 
					 dest_net, dest_node, LocalAddress.NetNum, LocalAddress.NodeNum );
			/*
			** Create the data packet and load it with data
			*/
			pkptr = op_pk_create_fmt("Unicast_Appl");
			op_pk_nfd_set(pkptr, "UniData", &DataSent, op_prg_mem_alloc,
						  op_prg_mem_free, sizeof(App_Data) );
			
			data_len = op_pk_total_size_get( pkptr )/8;
			/*
			** Encapsulate the data packet into the IP datagram by filling the sender address, destination 
			** address, data length, protocol Id and time-to-live
			*/
			ip_pkptr = op_pk_create_fmt("ip_dgram");
			op_pk_nfd_set(ip_pkptr, "data", pkptr );
			op_pk_nfd_set(ip_pkptr, "prot_id", UDP_PROTOCOL_ID );
			op_pk_bulk_size_set(ip_pkptr, data_len*8 );
			op_pk_nfd_set(ip_pkptr, "dest_net", dest_net );
			op_pk_nfd_set(ip_pkptr, "dest_node", dest_node );
			op_pk_nfd_set(ip_pkptr, "src_net", LocalAddress.NetNum );
			op_pk_nfd_set(ip_pkptr, "src_node", LocalAddress.NodeNum );
			op_pk_nfd_set(ip_pkptr, "orig_len", data_len );
			op_pk_nfd_set(ip_pkptr, "frag_len", data_len );
			op_pk_nfd_set(ip_pkptr, "ttl", 2 );
			op_pk_nfd_set(ip_pkptr, "frag", 0 );
			
			op_pk_send( ip_pkptr, IP_STREAM );      /*          Send the data packet to the IP module                */
			
			/*
			** Write the no. of data packets sent into the probe file 
			*/
			PROBE_ON
			{
				NoOfPktsSent++;
				op_stat_local_write( CUMM_PKT_SENT_STAT, NoOfPktsSent );
			}
			
			/*
			** Schedule the interrupt to send the data packet again
			*/
			op_intrpt_schedule_self( (op_sim_time() + (data_iat)), send_data_intr );
			}


		/** state (data) exit executives **/
		FSM_STATE_EXIT_FORCED (3, state3_exit_exec, "data")
			{
			}


		/** state (data) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		}


	FSM_EXIT (0,UniApp)
	}




void
UniApp_svar (prs_ptr,var_name,var_p_ptr)
	UniApp_state		*prs_ptr;
	char			*var_name, **var_p_ptr;
	{

	FIN (UniApp_svar (prs_ptr))

	*var_p_ptr = VOS_NIL;
	if (Vos_String_Equal ("my_objid" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_my_objid);
	if (Vos_String_Equal ("ip_objid" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_ip_objid);
	if (Vos_String_Equal ("data_iat" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_data_iat);
	if (Vos_String_Equal ("print_time" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_print_time);
	if (Vos_String_Equal ("NoOfPktsRecv" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfPktsRecv);
	if (Vos_String_Equal ("NoOfPktsSent" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfPktsSent);
	if (Vos_String_Equal ("MAX_NO_NETS" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_MAX_NO_NETS);
	if (Vos_String_Equal ("MAX_NO_NODES" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_MAX_NO_NODES);
	if (Vos_String_Equal ("DataSendInitTime" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_DataSendInitTime);
	if (Vos_String_Equal ("net_dist" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_dist);
	if (Vos_String_Equal ("node_dist" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_dist);
	if (Vos_String_Equal ("LocalAddress" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_LocalAddress);

	FOUT;
	}




void
UniApp_diag ()
	{
	double thru, thru1, tmpDataRec, tmpDataSend, tmp_sim_time, thru2, thru3;
	Packet*  pkptr;
	int in_strm,da1,da2,da3;
	FILE*  fil;
	

	FIN (UniApp_diag ())


	FOUT;
	}




void
UniApp_terminate ()
	{
	double thru, thru1, tmpDataRec, tmpDataSend, tmp_sim_time, thru2, thru3;
	Packet*  pkptr;
	int in_strm,da1,da2,da3;
	FILE*  fil;
	

	FIN (UniApp_terminate ())


	FOUT;
	}



Compcode
UniApp_init (pr_state_pptr)
	UniApp_state			**pr_state_pptr;
	{
	static VosT_Cm_Obtype	obtype = OPC_NIL;

	FIN (UniApp_init (pr_state_pptr))

	if (obtype == OPC_NIL)
		{
		if (Vos_Catmem_Register ("proc state vars (UniApp)",
			sizeof (UniApp_state), Vos_Nop, &obtype) == VOSC_FAILURE)
			FRET (OPC_COMPCODE_FAILURE)
		}

	if ((*pr_state_pptr = (UniApp_state*) Vos_Catmem_Alloc (obtype, 1)) == OPC_NIL)
		FRET (OPC_COMPCODE_FAILURE)
	else
		{
		(*pr_state_pptr)->current_block = 0;
		FRET (OPC_COMPCODE_SUCCESS)
		}
	}



